import dataclasses
import typing as t


@dataclasses.dataclass
class Document:
    id: str
    content: t.Dict[t.Any, t.Any]
    metadata: t.Optional[t.Dict[t.Any, t.Any]] = None


def parse_document(result: t.Dict[t.Any, t.Any]) -> Document:
    return Document(
        id=result["id"],
        content=result["content"],
        metadata=result.get("metadata"),
    )


UpsertDocumentT = t.Union[
    Document,
    t.Dict[str, t.Any],
    t.Tuple[t.Any, ...],
]


@dataclasses.dataclass
class DocumentScore:
    id: str
    score: float
    content: t.Dict[t.Any, t.Any]
    metadata: t.Optional[t.Dict[t.Any, t.Any]] = None


def parse_document_score(result: t.Dict[t.Any, t.Any]) -> DocumentScore:
    return DocumentScore(
        id=result["id"],
        score=result["score"],
        content=result["content"],
        metadata=result.get("metadata"),
    )


@dataclasses.dataclass
class RangeDocuments:
    next_cursor: str
    documents: t.List[Document]


def parse_range_documents(result: t.Dict[t.Any, t.Any]) -> RangeDocuments:
    return RangeDocuments(
        next_cursor=result["nextCursor"],
        documents=[parse_document(doc) for doc in result["vectors"]],
    )


@dataclasses.dataclass
class IndexInfo:
    document_count: int
    pending_document_count: int


def parse_index_info(result: t.Dict[t.Any, t.Any]) -> IndexInfo:
    return IndexInfo(
        document_count=result["vectorCount"],
        pending_document_count=result["pendingVectorCount"],
    )


@dataclasses.dataclass
class Info:
    document_count: int
    pending_document_count: int
    disk_size: int
    indexes: t.Dict[str, IndexInfo]


def parse_info(result: t.Dict[t.Any, t.Any]) -> Info:
    return Info(
        document_count=result["vectorCount"],
        pending_document_count=result["pendingVectorCount"],
        disk_size=result["indexSize"],
        indexes={
            index: parse_index_info(index_info)
            for index, index_info in result["namespaces"].items()
        },
    )


def parse_deleted(result: t.Dict[t.Any, t.Any]) -> int:
    return result["deleted"]  # type: ignore[no-any-return]



__version__ = "0.1.1"

from upstash_search.asyncio.index import AsyncIndex
from upstash_search.asyncio.search import AsyncSearch
from upstash_search.index import Index
from upstash_search.search import Search

__all__ = ["Search", "Index", "AsyncSearch", "AsyncIndex"]



import typing as t

from upstash_search.asyncio.http import AsyncRequester
from upstash_search.paths import (
    UPSERT_PATH,
    SEARCH_PATH,
    FETCH_PATH,
    DELETE_PATH,
    RANGE_PATH,
    RESET_PATH,
)
from upstash_search.types import (
    DocumentScore,
    Document,
    RangeDocuments,
    parse_document_score,
    parse_document,
    parse_deleted,
    parse_range_documents,
    UpsertDocumentT,
)
from upstash_search.utils import documents_to_payload


class AsyncIndex:
    """
    Represents an index of a database.

    Each index is an isolated component of a database where
    documents can be added, retrieved, searched, and deleted.
    """

    def __init__(
        self,
        name: str,
        requester: AsyncRequester,
    ):
        self._name = name
        self._requester = requester

    async def upsert(
        self,
        documents: t.Union[UpsertDocumentT, t.List[UpsertDocumentT]],
    ) -> None:
        """
        Upserts(updates or inserts) documents.

        Documents are identified by their unique ids.
        If a document with the same id already exists, it
        will be updated. Otherwise, a new document will be
        inserted.

        :param documents: Documents to upsert.
        """
        if not isinstance(documents, list):
            documents = [documents]

        payload = documents_to_payload(documents)

        await self._requester.post(
            path=UPSERT_PATH,
            payload=payload,
            index=self._name,
        )

    async def search(
        self,
        query: str,
        *,
        limit: int = 10,
        filter: str = "",
        reranking: bool = False,
        semantic_weight: float = 0.75,
        input_enrichment: bool = True,
    ) -> t.List[DocumentScore]:
        """
        Searches for documents matching the given query text.

        :param query: Query text to search for.
        :param limit: Number of documents to return.
        :param filter: Content filter to narrow down results.
        :param reranking: Optional boolean to use enhanced search result reranking. It will have additional
            cost when enabled. See [Search Pricing](https://upstash.com/pricing/search) for more details.
            False by default.
        :param semanticWeight: Optional relevance balance between semantic and keyword search (0-1 range, defaults to 0.75).
            For instance, 0.2 applies 20% semantic matching with 80% full-text matching.
            You can learn more about how Upstash Search works from [our docs](https://upstash.com/docs/search/features/algorithm).
        :param inputEnrichment: Optional boolean to enhance queries before searching (enabled by default).
        """

        payload = {
            "query": query,
            "topK": limit,
            "filter": filter,
            "reranking": reranking,
            "includeData": True,
            "includeMetadata": True,
            "semanticWeight": semantic_weight,
            "inputEnrichment": input_enrichment,
        }

        result = await self._requester.post(
            path=SEARCH_PATH,
            payload=payload,
            index=self._name,
        )

        document_scores = [
            parse_document_score(document_score) for document_score in result
        ]
        return document_scores

    async def fetch(
        self,
        *,
        ids: t.Optional[t.Sequence[str]] = None,
        prefix: t.Optional[str] = None,
    ) -> t.List[t.Optional[Document]]:
        """
        Fetches documents for the given ids or id prefix.

        :param ids: List of document ids to fetch.
        :param prefix: Prefix of the document ids to fetch.
        """

        payload: t.Dict[str, t.Any] = {
            "includeData": True,
            "includeMetadata": True,
        }

        if ids is not None:
            payload["ids"] = ids

        if prefix is not None:
            payload["prefix"] = prefix

        result = await self._requester.post(
            path=FETCH_PATH,
            payload=payload,
            index=self._name,
        )

        documents = [parse_document(doc) if doc is not None else None for doc in result]
        return documents

    async def delete(
        self,
        *,
        ids: t.Optional[t.Sequence[str]] = None,
        prefix: t.Optional[str] = None,
        filter: t.Optional[str] = None,
    ) -> int:
        """
        Deletes the documents having the given ids,
        id prefix, or content matching with filter.

        Returns how many documents are deleted.

        :param ids: List of document ids to delete.
        :param prefix: Prefix of the document ids to delete.
        :param filter: Filter to delete documents with matching content.
        """

        payload: t.Dict[str, t.Any] = {}
        if ids is not None:
            payload["ids"] = ids

        if prefix is not None:
            payload["prefix"] = prefix

        if filter is not None:
            payload["filter"] = filter

        result = await self._requester.post(
            path=DELETE_PATH,
            payload=payload,
            index=self._name,
        )

        deleted = parse_deleted(result)
        return deleted

    async def range(
        self,
        *,
        cursor: str = "",
        limit: int = 1,
        prefix: t.Optional[str] = None,
    ) -> RangeDocuments:
        """
        Ranges over the documents, starting from the cursor,
        and returning at most limit many documents.

        It can be used for paginating the documents, using
        the next cursor to use from the return value.

        :param cursor: Cursor to start range from.
        :param limit: At most how many documents to return.
        :param prefix: Optional document id prefix to range over.
        """

        payload = {
            "cursor": cursor,
            "limit": limit,
            "includeData": True,
            "includeMetadata": True,
        }

        if prefix is not None:
            payload["prefix"] = prefix

        result = await self._requester.post(
            path=RANGE_PATH,
            payload=payload,
            index=self._name,
        )

        range_documents = parse_range_documents(result)
        return range_documents

    async def reset(self) -> None:
        """
        Deletes all documents from the index
        and resets it to its initial state.
        """

        await self._requester.post(
            path=RESET_PATH,
            payload=None,
            index=self._name,
        )



import os
import platform as p
import time
import typing as t

import httpx

from upstash_search import __version__
from upstash_search.errors import UpstashError


def generate_headers(token: str, allow_telemetry: bool) -> t.Dict[str, str]:
    headers = {
        "Authorization": f"Bearer {token}",
    }

    if allow_telemetry:
        headers["Upstash-Telemetry-Sdk"] = f"upstash-search-py@v{__version__}"
        headers["Upstash-Telemetry-Runtime"] = f"python@v{p.python_version()}"

        if os.getenv("VERCEL"):
            platform = "vercel"
        elif os.getenv("AWS_REGION"):
            platform = "aws"
        else:
            platform = "unknown"

        headers["Upstash-Telemetry-Platform"] = platform

    return headers


class Requester:
    def __init__(
        self,
        url: str,
        token: str,
        retries: int,
        retry_interval: float,
        allow_telemetry: bool,
    ):
        self._client = httpx.Client(
            timeout=httpx.Timeout(
                timeout=600.0,
                connect=10.0,
            )
        )
        self._url = url
        self._headers = generate_headers(token, allow_telemetry)
        self._retries = retries
        self._retry_interval = retry_interval

    def post(
        self,
        path: str,
        payload: t.Optional[t.Any] = None,
        index: t.Optional[str] = None,
    ) -> t.Any:
        if index:
            url = f"{self._url}{path}/{index}"
        else:
            url = f"{self._url}{path}"

        response = None
        last_error = None

        for attempts_left in range(max(0, self._retries), -1, -1):
            try:
                response = self._client.post(
                    url=url,
                    headers=self._headers,
                    json=payload,
                )
                break

            except Exception as e:
                last_error = e
                if attempts_left > 0:
                    time.sleep(self._retry_interval)

        if response is None:
            assert last_error is not None
            raise last_error

        body = response.json()
        if "error" in body:
            raise UpstashError(body["error"])

        return body["result"]



LIST_INDEXES_PATH = "/list-namespaces"
DELETE_INDEX_PATH = "/delete-namespace"
INFO_PATH = "/info"
UPSERT_PATH = "/upsert-data"
SEARCH_PATH = "/search"
FETCH_PATH = "/fetch"
DELETE_PATH = "/delete"
RANGE_PATH = "/range"
RESET_PATH = "/reset"



import os
import typing as t

from upstash_search.asyncio.http import AsyncRequester
from upstash_search.asyncio.index import AsyncIndex
from upstash_search.paths import LIST_INDEXES_PATH, DELETE_INDEX_PATH, INFO_PATH
from upstash_search.types import Info, parse_info


class AsyncSearch:
    """
    Upstash Search client that uses the HTTP API to perform operations.

    ```python
    from upstash_search import AsyncSearch

    client = AsyncSearch(
        url="UPSTASH_SEARCH_REST_URL",
        token="UPSTASH_SEARCH_REST_TOKEN",
    )
    index = client.index("INDEX_NAME")
    ```
    """

    def __init__(
        self,
        url: str,
        token: str,
        *,
        retries: int = 3,
        retry_interval: float = 1.0,
        allow_telemetry: bool = True,
    ):
        self._url = url
        self._requester = AsyncRequester(
            url=url,
            token=token,
            retries=retries,
            retry_interval=retry_interval,
            allow_telemetry=allow_telemetry,
        )

    def index(self, name: str) -> AsyncIndex:
        """
        Returns an index for the given name.

        Each index is an isolated component of a database where
        documents can be added, retrieved, searched, and deleted.

        :param name: Name of the index.
        """

        return AsyncIndex(name, self._requester)

    async def list_indexes(self) -> t.List[str]:
        """
        Returns the names of the indexes of the database.
        """

        indexes = await self._requester.post(
            path=LIST_INDEXES_PATH,
        )
        return indexes  # type: ignore[no-any-return]

    async def delete_index(self, name: str) -> None:
        """
        Deletes the given index of the database.

        :param name: Name of the index to delete.
        """

        await self._requester.post(
            path=DELETE_INDEX_PATH,
            index=name,
        )

    async def info(self) -> Info:
        """
        Returns the database info.
        """

        result = await self._requester.post(
            path=INFO_PATH,
        )

        return parse_info(result)

    @classmethod
    def from_env(
        cls,
        *,
        retries: int = 3,
        retry_interval: float = 1.0,
        allow_telemetry: bool = True,
    ) -> "AsyncSearch":
        """
        Load the credentials from environment variables,
        and returns a client.
        """

        return cls(
            url=os.environ["UPSTASH_SEARCH_REST_URL"],
            token=os.environ["UPSTASH_SEARCH_REST_TOKEN"],
            retries=retries,
            retry_interval=retry_interval,
            allow_telemetry=allow_telemetry,
        )



import typing as t

from upstash_search.errors import ClientError
from upstash_search.types import Document


def documents_to_payload(
    documents: t.Sequence[t.Union[t.Dict[t.Any, t.Any], t.Tuple[t.Any, ...], Document]],
) -> t.List[t.Dict[str, t.Any]]:
    payload = []
    for doc in documents:
        parsed_doc = _parse_document(doc)
        payload.append(
            {
                "id": parsed_doc.id,
                "content": parsed_doc.content,
                "metadata": parsed_doc.metadata,
            }
        )

    return payload


def _parse_document(
    document: t.Union[t.Dict[t.Any, t.Any], t.Tuple[t.Any, ...], Document],
) -> Document:
    if isinstance(document, Document):
        return document
    elif isinstance(document, tuple):
        return _parse_tuple(document)
    elif isinstance(document, dict):
        return _parse_dict(document)
    else:
        raise ClientError(f"Unsupported document type: {document}")


def _parse_tuple(document: t.Tuple[t.Any, ...]) -> Document:
    if len(document) < 2:
        raise ClientError(
            "The tuple must contain at least two elements; "
            "one for id, and other for content."
        )

    doc_id = document[0]
    content = document[1]

    if len(document) > 2:
        metadata = document[2]
    else:
        metadata = None

    return Document(
        id=doc_id,
        content=content,
        metadata=metadata,
    )


def _parse_dict(document: t.Dict[t.Any, t.Any]) -> Document:
    doc_id = document["id"]
    content = document["content"]
    metadata = document.get("metadata")

    return Document(
        id=doc_id,
        content=content,
        metadata=metadata,
    )
